﻿/**
Filename: GUD900_Serial_Async.cpp
Purpose: Establish a communication interface between an
		 Arduino Uno board and a Noritake GUD display 
		 module through Asynchronous Serial communication. 
		 This interface features both read and write functionality.
@author: Cody Johnson
Email: cody.ele@noritake.com
Note: SoftwareSerial.cpp was heavily referenced while
	  writing this interface. Additionally, PIN7 must be tied to 
	  GND in order for read to work.
*/

#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <Arduino.h>
#include <util/delay_basic.h>
#include <GUD900_Interface.h>
#include "GUD900_Serial_Async.h"
	
GUD900_Serial_Async *GUD900_Serial_Async::active_object = 0;
char GUD900_Serial_Async::_receive_buffer[_SS_MAX_RX_BUFF];
volatile uint8_t GUD900_Serial_Async::_receive_buffer_tail = 0;
volatile uint8_t GUD900_Serial_Async::_receive_buffer_head = 0;

// private static method for timing
inline void GUD900_Serial_Async::tunedDelay(uint16_t delay)
{
	_delay_loop_2(delay);
}

//Helper method to aid in delay calculations.
uint16_t GUD900_Serial_Async::subtract_cap(uint16_t num, uint16_t sub) 
{
	if (num > sub)
		return num - sub;
	else
		return 1;
}

/**
The Asynchronous Serial Communication constructor.

@param  baudrate The desired baudrate for communication.
@param  out      The data out pin number being used on the Arduino Uno.
@param  in       The data in pin number being used on the Arduino Uno.
@param  busy     The busy pin number being used on the Arduino Uno.
@param  reset    The reset pin number being used on the Arduino Uno.
@return none
*/
GUD900_Serial_Async::GUD900_Serial_Async(long baudRate, uint8_t out, uint8_t in, uint8_t busy, uint8_t reset):
OUT_PIN(out), BUSY_PIN(busy), RESET_PIN(reset), IN_PIN(in), _rx_delay_centering(0),
_rx_delay_intrabit(0), _rx_delay_stopbit(0), _tx_delay(0)
{
	_rx_delay_centering = _rx_delay_intrabit = _rx_delay_stopbit = _tx_delay = 0;
	
	/**Set TX*/
	_transmitBitMask = digitalPinToBitMask(out);
	uint8_t port = digitalPinToPort(out);
	_transmitPortRegister = portOutputRegister(port);
	
	//Set RX
	_receiveBitMask = digitalPinToBitMask(in);
	uint8_t portR = digitalPinToPort(in);
	_receivePortRegister = portInputRegister(portR);

	//Updated Calculations from Arduino SofwareSerial
	uint16_t bit_delay = (F_CPU / baudRate) / 4;
	_tx_delay = subtract_cap(bit_delay, 15 / 4);

	if(digitalPinToPCICR(out))
	{
		#if GCC_VERSION > 40800
		_rx_delay_centering = subtract_cap(bit_delay / 2, (4 + 4 + 75 + 17 - 23) / 4);
		_rx_delay_intrabit = subtract_cap(bit_delay, 23 / 4);
		_rx_delay_stopbit = subtract_cap(bit_delay * 3 / 4, (37 + 11) / 4);
		
		#else
		_rx_delay_centering = subtract_cap(bit_delay / 2, (4 + 4 + 97 + 29 - 11) / 4);
		_rx_delay_intrabit = subtract_cap(bit_delay, 11 / 4);
		_rx_delay_stopbit = subtract_cap(bit_delay * 3 / 4, (44 + 17) / 4);
		
		#endif
		
		*digitalPinToPCICR(in) |= _BV(digitalPinToPCICRbit(in));
		_pcint_maskreg = digitalPinToPCMSK(in);
		_pcint_maskvalue = _BV(digitalPinToPCMSKbit(in));
			
		tunedDelay(_tx_delay);
	}
	listen();
}

/**
Initializes the Asynchronous Serial protocol.

@return none
*/
void GUD900_Serial_Async::init()
{
	RAISE(RESET);
	DIRECTION(OUT,1);
	DIRECTION(RESET,1);
	DIRECTION(BUSY,0);
	DIRECTION(IN, 0);
	//set model class
	//if(getModelClass==7003 || getModelClass==7903 || getModelClass==7053) {
	_inverse_logic = 0; //might need to switch to different logic if nothing works
	RAISE(OUT);
	//}else{
	//_inverse_logic = 1;
	//LOWER(OUT);
	//}
}

/**
Write function for 1 bit of data.

@param  pin_state The desired pin state to write. Either HIGH or LOW.
@return none
*/
void GUD900_Serial_Async::tx_pin_write(uint8_t pin_state)
{
	if (pin_state == LOW)
		*_transmitPortRegister &= ~_transmitBitMask;
	else
		*_transmitPortRegister |= _transmitBitMask;
}

/**
Reads the RX pin state.

@return The RX pin state.
*/
uint8_t GUD900_Serial_Async::rx_pin_read()
{
	return *_receivePortRegister & _receiveBitMask;
}

/**
Write function for 1 byte of data.

@param  data The data to be written.
@return none
*/
void GUD900_Serial_Async::write(uint8_t data)
{
	//Local variables to help ensure the best timing possible for transmit.
	bool inv = _inverse_logic;
	uint16_t delay = _tx_delay;
	volatile uint8_t *reg = _transmitPortRegister;
	uint8_t reg_mask = _transmitBitMask;
	uint8_t inv_mask = ~_transmitBitMask;
	uint8_t oldSREG = SREG;

	// Check for busy signal
	if (!inv)
		while (CHECK(BUSY));
	else
		while (!CHECK(BUSY));
	
	if(inv){
		data = ~data;
	}
	
	cli();  // turn off interrupts for a clean transmit

	// Write the start bit
	if (inv)
		*reg |= reg_mask;
	else
		*reg &= inv_mask;
	
	tunedDelay(delay);
	
	// Write each of the 8 bits
	for (uint8_t i = 8; i > 0; --i) {
		
		if (data & 1)
			*reg |= reg_mask; // send 1
		else
			*reg &= inv_mask; // send 0
		
		tunedDelay(delay);
		data >>= 1;
	}

	if (inv)
		*reg &= inv_mask;
	else
		*reg |= reg_mask;

	tunedDelay(delay);
	SREG = oldSREG; // turn interrupts back on
}

/**
The receive routine called by the interrupt handler

@return none
*/
void GUD900_Serial_Async::recv()
{
	#if GCC_VERSION < 40302
	// Work-around for avr-gcc 4.3.0 OSX version bug
	// Preserve the registers that the compiler misses
	// (courtesy of Arduino forum user *etracer*)
	asm volatile(
		"push r18 \n\t"
		"push r19 \n\t"
		"push r20 \n\t"
		"push r21 \n\t"
		"push r22 \n\t"
		"push r23 \n\t"
		"push r26 \n\t"
		"push r27 \n\t"
	::);
	#endif

	uint8_t d = 0;

	// If RX line is high, then we don't see any start bit
	// so interrupt is probably not for us
	if (_inverse_logic ? rx_pin_read() : !rx_pin_read())
	{
		// Disable further interrupts during reception, this prevents
		// triggering another interrupt directly after we return, which can
		// cause problems at higher baud rates.
		setRxIntMsk(false);

		// Wait approximately 1/2 of a bit width to "center" the sample
		tunedDelay(_rx_delay_centering);
		//DebugPulse(_DEBUG_PIN2, 1);

		// Read each of the 8 bits
		for (uint8_t i=8; i > 0; --i)
		{
			tunedDelay(_rx_delay_intrabit);
			d >>= 1;
			if (rx_pin_read())
			d |= 0x80;
		}

		if (_inverse_logic)
			d = ~d;

		// if buffer full, set the overflow flag and return
		uint8_t next = (_receive_buffer_tail + 1) % _SS_MAX_RX_BUFF;
		if (next != _receive_buffer_head)
		{
			// save new data in buffer: tail points to where byte goes
			_receive_buffer[_receive_buffer_tail] = d; // save new byte
			_receive_buffer_tail = next;
		}
		else
		{
			_buffer_overflow = true;
		}

		// skip the stop bit
		tunedDelay(_rx_delay_stopbit);

		// Re-enable interrupts when we're sure to be inside the stop bit
		setRxIntMsk(true);
	}

	#if GCC_VERSION < 40302
	// Work-around for avr-gcc 4.3.0 OSX version bug
	// Restore the registers that the compiler misses
	asm volatile(
		"pop r27 \n\t"
		"pop r26 \n\t"
		"pop r23 \n\t"
		"pop r22 \n\t"
		"pop r21 \n\t"
		"pop r20 \n\t"
		"pop r19 \n\t"
		"pop r18 \n\t"
	::);
	#endif
}

/**
Sets the receive interrupt mask.

@param  enable ENABLE (logic 1) or DISABLE (logic 0) the receive interrupt mask.
@return none
*/
void GUD900_Serial_Async::setRxIntMsk(bool enable)
{
	if (enable)
		*_pcint_maskreg |= _pcint_maskvalue;
	else
		*_pcint_maskreg &= ~_pcint_maskvalue;
}

/**
Reads data from the receive buffer.

IMPORTANT NOTE: PIN7 of the GUD module must be
tied to ground in order to use interrupts and
enable read functionality.

@return If -1, the bus is not listening.
		If -2, the buffer is empty.
		Otherwise, the read data is returned.
*/
int GUD900_Serial_Async::read()
{
	if (!isListening())
		return -1;

	// Empty buffer?
	if (_receive_buffer_head == _receive_buffer_tail)
		return -2;

	// Read from "head"
	uint8_t d = _receive_buffer[_receive_buffer_head]; // grab next byte
	_receive_buffer_head = (_receive_buffer_head + 1) % _SS_MAX_RX_BUFF;
	return d;
}

/**
Determine if the device is available to be read.

@return 0 if unavailable
		1 if available
*/
int GUD900_Serial_Async::available()
{
	if (!isListening())
		return 0;

	return (_receive_buffer_tail + _SS_MAX_RX_BUFF - _receive_buffer_head) % _SS_MAX_RX_BUFF;
}

/**
Determine if the device is listening and active.

@return 0 if not active
		1 if active
*/
bool GUD900_Serial_Async::isListening()
{
	return this == active_object;
}

/**
This function sets the current object as the "listening"
one and returns true if it replaces another.

@return 1 if the listening object replaced other object.
		0 otherwise
*/
bool GUD900_Serial_Async::listen()
{
	if (!_rx_delay_stopbit)
		return false;

	if (active_object != this)
	{
		if (active_object)
		active_object->stopListening();

		_buffer_overflow = false;
		_receive_buffer_head = _receive_buffer_tail = 0;
		active_object = this;

		setRxIntMsk(true);
		return true;
	}

	return false;
}

/**
Stop listening. Returns true if we were actually listening.

@return 1 if we were actually listening
		0 otherwise
*/
bool GUD900_Serial_Async::stopListening()
{
	if (active_object == this)
	{
		setRxIntMsk(false);
		active_object = NULL;
		return true;
	}
	return false;
}

/**
Function to reset the GUD module.

@return none
*/
void GUD900_Serial_Async::hardReset(){
	init();
	LOWER(RESET);
	_delay_ms(1);
	RAISE(RESET);
	_delay_ms(100);
	while (CHECK(BUSY));
}

/** static */
inline void GUD900_Serial_Async::handle_interrupt()
{
	if (active_object)
	{
		active_object->recv();
	}
}

#if defined(PCINT0_vect)
ISR(PCINT0_vect)
{
	GUD900_Serial_Async::handle_interrupt();
}
#endif

#if defined(PCINT1_vect)
ISR(PCINT1_vect, ISR_ALIASOF(PCINT0_vect));
#endif

#if defined(PCINT2_vect)
ISR(PCINT2_vect, ISR_ALIASOF(PCINT0_vect));
#endif

#if defined(PCINT3_vect)
ISR(PCINT3_vect, ISR_ALIASOF(PCINT0_vect));
#endif