/**
Filename: GUD900_Serial_SPI.h
Purpose: Establish a communication interface between an
		 Arduino Uno board and a Noritake GUD display 
		 module through SPI. This interface features both
		 read and write functionality.
Author: Cody Johnson
Email: cody.ele@noritake.com
Note: GU7000_Serial_SPI.h was heavily referenced while
	  writing this interface.
*/

//SPI command definitions.
#define DATA_WRITE_CMD 0x44
#define DATA_READ_CMD 0x54
#define STATUS_READ_CMD 0x58

class GUD900_Serial_SPI : public GUD900_Interface{

protected:
	//Pin definitions
	unsigned OUT_PIN:4;
	unsigned IN_PIN:4;
	unsigned BUSY_PIN:4;
	unsigned SCK_PIN:4;
	unsigned RESET_PIN:4;
	unsigned CS_PIN:4;

public:

	/**
	SPI Communication Constructor
	
	@param  out   The data out pin number being used on the Arduino Uno.
	@param  in    The data in pin number being used on the Arduino Uno.
	@param  busy  The busy pin number being used on the Arduino Uno.
	@param  sck   The serial clock pin number being used on the Arduino Uno.
	@param  reset The reset pin number being used on the Arduino Uno.
	@param  cs    The chip select pin number being used on the Arduino Uno.
	#return none
	*/
	GUD900_Serial_SPI(uint8_t out, uint8_t in, uint8_t busy, uint8_t sck, uint8_t reset, uint8_t cs):
		OUT_PIN(out), BUSY_PIN(busy), SCK_PIN(sck), RESET_PIN(reset), CS_PIN(cs), IN_PIN(in)
	{
	}
	
	/**
	Initializes the SPI protocol.
	
	@return none
	*/
	void init()
	{
		LOWER(OUT);
    	RAISE(SCK);
        RAISE(RESET);
        RAISE(CS);
    	DIRECTION(CS,1);
    	DIRECTION(RESET,1);
    	DIRECTION(OUT,1);
    	DIRECTION(SCK,1);
    	DIRECTION(BUSY,0);
		DIRECTION(IN, 0);
	}

	/**
	Write function for 1 byte of data.
	
	@param  data The data to be written to the display.
	@return none
	*/
	void write(uint8_t data)
	{
		LOWER(CS);
        while (CHECK(BUSY));
		//Send write command
		for (uint8_t i = 0x80; i; i >>= 1) {
    		LOWER(SCK);
    		_delay_us(1);
    		if (DATA_WRITE_CMD & i) RAISE(OUT); else LOWER(OUT);
    		RAISE(SCK);
    		_delay_us(1);
    	} 
		//Send data byte
    	for (uint8_t i = 0x80; i; i >>= 1) {
    		LOWER(SCK);
    		_delay_us(1);
    		if (data & i) RAISE(OUT); else LOWER(OUT);
    		RAISE(SCK);
    		_delay_us(1);
    	}
     	_delay_us(17);
        RAISE(CS); 
	}
	
	/**
	Function to read the number of available 
	bytes to read from the module. 
	
	@return The number of bytes available to be read.
	*/
	uint8_t statusRead()
	{
		uint8_t value = 0;
		LOWER(CS);
        while (CHECK(BUSY));
		//Send read command
		for (uint8_t i = 0x80; i; i >>= 1) {
    		LOWER(SCK);
    		_delay_us(1);
    		if (STATUS_READ_CMD & i) RAISE(OUT); else LOWER(OUT);
    		RAISE(SCK);
    		_delay_us(1);
    	}
		_delay_us(17);
		value = read();
		finishRead();
		return value;
	}
	
	/**
	Function to start reading data from the module.
	
	@return none
	*/
	void startRead()
	{
		LOWER(CS);
        while (CHECK(BUSY));
		//Send read command
		for (uint8_t i = 0x80; i; i >>= 1) {
    		LOWER(SCK);
    		_delay_us(1);
    		if (DATA_READ_CMD & i) RAISE(OUT); else LOWER(OUT);
    		RAISE(SCK);
    		_delay_us(1);
    	}
		_delay_us(17);
	}
	
	/**
	Read function for 1 byte of data.
	
	@return The byte read from the display.
	*/
	uint8_t read()
	{
		uint8_t value = 0;
		for(int8_t i = 0; i < 8; i++){
			//Shift for incoming bit
			value <<= 1;
			LOWER(SCK);
			_delay_us(1);
			//Store incoming bit
			if(CHECK(IN)){
				value += 1;
			}else{
				value += 0;
			}
			RAISE(SCK);
			_delay_us(1);
		}
		return value;
	}
	
	/**
	Finish reading data by delaying SCK
	and raising ChipSelect.
	
	@return none
	*/
	void finishRead()
	{
		_delay_us(17);
		RAISE(CS);
	}
	
	/**
	Function to reset the module.
	
	@return none
	*/
	void hardReset()
	{
		init();
        LOWER(CS);
        LOWER(RESET);
    	_delay_ms(1);
    	RAISE(RESET);
        _delay_ms(100);
    	RAISE(CS);
	}
};