/**
Filename: GUD900_TWI.h
Purpose: Establish a communication interface between and
		 Arduino Uno board and a Noritake GUD display 
		 module through TWI/I2C. This interface features both
		 read and write functionality.
Author: Cody Johnson
Email: cody.ele@noritake.com
Note: The Arduino's Wire.cpp was heavily referenced while
	  writing this interface.
*/

//Include important header files from C
extern "C" {
	#include <stdlib.h>
	#include <string.h>
	#include <inttypes.h>
	#include "utility\twi.h"
}

#define BUFFER_LENGTH 32

class GUD900_TWI : public GUD900_Interface{

private:
	//Slave address for GUD900
	uint8_t GUD900_slaveAddress;
	uint32_t clockFrequency;
	//Initialize rx buffer
	uint8_t rxBuffer[BUFFER_LENGTH];
	uint8_t rxBufferIndex = 0;
	uint8_t rxBufferLength = 0;
	//Initialize tx buffer
	uint8_t txBuffer[BUFFER_LENGTH];
	uint8_t txBufferIndex = 0;
	uint8_t txBufferLength = 0;
	//Initialize transmitting flag
	uint8_t transmitting = 0;
	void (*user_onRequest)(void);
	void (*user_onReceive)(int);
	
protected:
	//Pin definitions
	unsigned MBUSY_PIN:4;
	unsigned TRDY_PIN:4;
	unsigned RESET_PIN:4;
	
public:
	/**
	I2C Interface Constructor
	
	@param  slaveAddress The slave address of the display.
	@param  clk_Freq     The desired clock frequency for communication.
	@param  mbusy        The module busy pin number being used on the Arduino Uno.
	@param  trdy         The transmit ready pin number being used on the Arduino Uno.
	@param  reset        The reset pin number being used on the Arduino Uno.
	@return none
	*/
	GUD900_TWI(uint8_t slaveAddress, uint32_t clk_Freq, uint8_t mbusy, uint8_t trdy, uint8_t reset):
		GUD900_slaveAddress(slaveAddress), clockFrequency(clk_Freq), MBUSY_PIN(mbusy), TRDY_PIN(trdy),
		RESET_PIN(reset)
	{
	}
	
	/**
	Initializes the I2C protocol.
	
	@return none
	*/
	void init()
	{
		RAISE(RESET);
		DIRECTION(RESET, 1);
		DIRECTION(MBUSY, 0);
		DIRECTION(TRDY, 0);
		twi_setAddress(GUD900_slaveAddress);

		//need to set a tx and rx buffer
		rxBufferIndex = 0;
		rxBufferLength = 0;
		txBufferIndex = 0;
		txBufferLength = 0;
		//Set the clock frequency
		TWBR = ((F_CPU / clockFrequency) - 16) / 2;
		twi_init();
	}
	
	/**
	Starts TWI transmission.
	
	@return none
	*/
	void start()
	{
		transmitting = 1;
		txBufferIndex = 0;
		txBufferLength = 0;
	}
	
	/**
	Write function for 1 byte of data.
	
	@param  data The data to be written to the display.
	@return none
	*/
	void write(uint8_t data)
	{
		start();
		if(transmitting){
			if(txBufferLength >= BUFFER_LENGTH){
				//setWriteError();
				return; 
			}
			txBuffer[txBufferIndex] = data;
			++txBufferIndex;
			txBufferLength = txBufferIndex;
		}else{
			twi_transmit(&data, 1);
		}
		stop(true);
		return;
	}
	
	/**
	Write function for multiple bytes of data.
	
	@param  *data    The data to be written to the display.
	@param  quantity The size of the data to be written.
	@return none
	*/
	void write(const uint8_t *data, size_t quantity)
	{
		start();
		if(transmitting){
			for(size_t i = 0; i < quantity; ++i){
				write(data[i]);
			}
		}else{
			twi_transmit(data, quantity);
		}
		stop(true);
		return;
	}
	
	/**
	Reads in one byte of data.
	
	must be called in:
	slave rx event callback
	or after requestFrom(address, numBytes)
	
	@return The byte read in from the slave device.
	*/
	int read()
	{
		int value = -1;
		// get each successive byte on each call
		if(rxBufferIndex < rxBufferLength){
			value = rxBuffer[rxBufferIndex];
			++rxBufferIndex;
		}

		return value;
	}
	
	/**
	Checks to see if the slave device is available.
	
	must be called in:
	slave rx event callback
	or after requestFrom(address, numBytes)
	
	@return 0 if unavailable
			non-zero if available
	*/
	int available()
	{
		return rxBufferLength - rxBufferIndex;
	}
	
	/**
	Makes a read request to a certain address for
	a specified amount of data bytes.
	
	@param  address  The address to send the request to.
	@param  quantity The number of data bytes requested.
	@param  sendStop Indicate whether more data is needed
					 after the transmission is completed.
	@return The data read.
	*/
	uint8_t requestFrom(uint8_t address, uint8_t quantity, uint8_t sendStop)
	{
		// clamp to buffer length
		if(quantity > BUFFER_LENGTH){
			quantity = BUFFER_LENGTH;
		}
		// perform blocking read into buffer
		uint8_t read = twi_readFrom(address, rxBuffer, quantity, sendStop);
		// set rx buffer iterator vars
		rxBufferIndex = 0;
		rxBufferLength = read;

		return read;
	}
	
	/**
	Stops the I2C communication.
	
	@param  stopFlag Indicate whether a stop is needed.
	@return The writeTo result.
	*/
	uint8_t stop(uint8_t stopFlag)
	{
		int8_t ret = twi_writeTo(GUD900_slaveAddress, txBuffer, txBufferLength, 1, stopFlag);
		txBufferIndex = 0;
		txBufferLength = 0;
		transmitting = 0;
		return ret;
	}
	
	//Behind the scenes function that is called when data is requested
	void onRequestService(void)
	{
		// don't bother if user hasn't registered a callback
		if(!user_onRequest){
			return;
		}
		// reset tx buffer iterator vars
		// !!! this will kill any pending pre-master sendTo() activity
		txBufferIndex = 0;
		txBufferLength = 0;
		// alert user program
		user_onRequest();
	}
	
	//Behind the scenes function that is called when data is requested
	void onReceiveService(uint8_t* inBytes, int numBytes)
	{
		// don't bother if user hasn't registered a callback
		if(!user_onReceive){
			return;
		}
		// don't bother if rx buffer is in use by a master requestFrom() op
		// i know this drops data, but it allows for slight stupidity
		// meaning, they may not have read all the master requestFrom() data yet
		if(rxBufferIndex < rxBufferLength){
			return;
		}
		// copy twi rx buffer into local read buffer
		// this enables new reads to happen in parallel
		for(uint8_t i = 0; i < numBytes; ++i){
			rxBuffer[i] = inBytes[i];    
		}
		// set rx iterator vars
		rxBufferIndex = 0;
		rxBufferLength = numBytes;
		// alert user program
		user_onReceive(numBytes);
	}
	
	/**
	Function to reset the module.
	
	@return none
	*/
	void hardReset()
	{
		LOWER(RESET);
		_delay_ms(1);
		RAISE(RESET);
		_delay_ms(100);
		while(CHECK(MBUSY));
		_delay_ms(1);
	}
};