/**
Filename: Noritake_VFD_GUD900.h
Purpose: Operate a GUD module through its established commands.
Author: Cody Johnson
Email: cody.ele@noritake.com
Note: Noritake_VFD_GU7000.h was heavily referenced while
	  writing this file.
*/

#ifndef NORITAKE_VFD_GUD900_H
#define NORITAKE_VFD_GUD900_H

#include <stdint.h>
#include <stddef.h>
#include <util/delay.h>

/**
The available image memory areas.
*/
enum ImageMemoryArea {
    FlashImageArea =  1,
    ScreenImageArea = 2
};

/**
The available scroll mode to be used.

WrappingMode: Wraps the existing text down to the next line
			  when text has gone beyond the boundaries of the
			  display.
VertScrollMode: Scrolls the existing text vertically when 
				text has gone beyond the boundaries of the display.
HorizScrollMode: Scroll the existing text vertically when text has
				 gone beyond the boundaries of the display.
*/
enum ScrollMode {
    WrappingMode =    1,
    VertScrollMode =  2,
    HorizScrollMode = 3
};

/**
The available composition modes to be used.
*/
enum CompositionMode {
    NormalCompositionMode = 0,
    OrCompositionMode =     1,
    AndCompositionMode =    2,
    XorCompositionMode =    3
};

/**
The available screensavers to be used.
*/
enum ScreenSaver {
    AllDotsOffSaver = 2,
    AllDotsOnSaver =  3,
    InvertSaver =     4
};

/**
The available LED backlight colors to be used.
NOTE: Can only be used for the 7040 display model.
*/
enum LEDColor {
    NoLight =       0x000,
    BlueLight =     0x00f,
    GreenLight =    0x0f0,
    CyanLight =     0x0ff,
    RedLight =      0xf00,
    MagentaLight =  0xf0f,
    SmokeLight =    0xfff
};

/**
The available font formats to be used.
*/
enum FontFormat {
    GUD9005x7Format = 0,
    GUD9007x8Format = 1,
    CUUFormat =       2,
    LCDFormat =       CUUFormat     
};



class Noritake_VFD_GUD900 {

    void initialState();
    void printNumber(unsigned long number, uint8_t base);
    void printNumber(unsigned x, uint8_t y, unsigned long number, uint8_t base);
    void command(uint8_t data);
    void us_command(uint8_t group, uint8_t cmd);
    void command(uint8_t prefix, uint8_t group, uint8_t cmd);
    void command_xy(unsigned x, unsigned y);
    void command_xy1(unsigned x, unsigned y);
    void crlf();
    
    GUD900_Interface *io;    

public:
	//The width of the display in dots.
    uint8_t width;
	//The height of the display in dots.
    uint8_t height;
	//The number of character lines available.
    uint8_t lines;
	//The model number of the display.
    unsigned modelClass;
	//
    bool generation;
	
	/**
	Public default constructor.
	*/
    Noritake_VFD_GUD900() {
        this->modelClass = 900;
		this->generation = false;
    }

	/**
	Display interface initializer.
	
	@param  &interface The interface to be used.
	@return none
	*/
    void interface(GUD900_Interface &interface) {
        io = &interface;
		//io->init();
    }
    
	/**
	Stores the height and width of the current display.
	
	@param  width  The width of the display in dots.
	@param  height The height of the display in dots.
	@return none
	*/
    void begin(uint8_t width, uint8_t height) {
        this->width = width;
        this->height = height;
        this->lines = this->height / 8;
    }
    
	/**
	Stores the class model number.
	
	@param   modelClass The model number of the current display.
	@return none
	*/
    void isModelClass(unsigned modelClass) {
        this->modelClass = modelClass;
		io->getModelClass = modelClass;
    }
	
	/**
	Stores the generation of the current display.
	
	@param c The generation character of the current display.
	@return none
	*/
	void isGeneration(char c){
		this->generation - toupper(c) == 'S';
	}

    void GUD900_back();
    void GUD900_forward();
    void GUD900_lineFeed();
    void GUD900_home();
    void GUD900_carriageReturn();
    void GUD900_setCursor(unsigned x, unsigned y);
    void GUD900_clearScreen();
    void GUD900_cursorOn();
    void GUD900_cursorOff();
    void GUD900_init();
    void GUD900_reset();
    void GUD900_useMultibyteChars(bool enable);
    void GUD900_setMultibyteCharset(uint8_t code);
    void GUD900_useCustomChars(bool enable);
    void GUD900_defineCustomChar(uint8_t code, FontFormat format, const uint8_t *data);
    void GUD900_deleteCustomChar(uint8_t code);
    void GUD900_setAsciiVariant(uint8_t code);
    void GUD900_setCharset(uint8_t code);
    void GUD900_setScrollMode(ScrollMode mode);
    void GUD900_setHorizScrollSpeed(uint8_t speed);
    void GUD900_invertOn();
    void GUD900_invertOff();
    void GUD900_setCompositionMode(CompositionMode mode);
    void GUD900_setScreenBrightness(unsigned level);
    void GUD900_wait(uint8_t time);
    void GUD900_scrollScreen(unsigned x, unsigned y, unsigned count, uint8_t speed);
    void GUD900_blinkScreen();
    void GUD900_blinkScreen(bool enable, bool reverse, uint8_t on, uint8_t off, uint8_t times);
    void GUD900_displayOn();
    void GUD900_displayOff();
    void GUD900_screenSaver(ScreenSaver mode);
    void GUD900_drawImage(unsigned width, uint8_t height, const uint8_t *data);
	void GUD900_setFROMImage(unsigned long address, unsigned long length, const uint8_t *data);
    void GUD900_drawFROMImage(unsigned long address, uint8_t srcHeight, unsigned width, uint8_t height);
    void GUD900_setFontStyle(bool proportional, bool evenSpacing);
    void GUD900_setFontSize(uint8_t x, uint8_t y, bool tall);
    void GUD900_selectWindow(uint8_t window);
    void GUD900_defineWindow(uint8_t window, unsigned x, unsigned y, unsigned width, unsigned height);
    void GUD900_deleteWindow(uint8_t window);
    void GUD900_joinScreens();
    void GUD900_separateScreens();
    void GUD900_setBacklightColor(uint8_t r, uint8_t g, uint8_t b);
    void GUD900_setBacklightColor(unsigned rgb);

	//Basic print functions
    void print(char c);
    void print(const char *str);
    void print(const char *buffer, size_t size);
    void print(int number, uint8_t base);
    void print(unsigned number, uint8_t base);
    void print(long number, uint8_t base);
    void print(unsigned long number, uint8_t base);
    void println(char c);
    void println(const char *str);
    void println(const char *buffer, size_t size);
    void println(int number, uint8_t base);
    void println(unsigned number, uint8_t base);
    void println(long number, uint8_t base);
    void println(unsigned long number, uint8_t base);

    void GUD900_fillRect(unsigned x0, unsigned y0, unsigned x1, unsigned y1, bool on = true);
    
    void print(unsigned x, uint8_t y, char c);
    void print(unsigned x, uint8_t y, const char *str);
    void print(unsigned x, uint8_t y, const char *buffer, uint8_t len);
    void print(unsigned x, uint8_t y, int number, uint8_t base);
    void print(unsigned x, uint8_t y, unsigned number, uint8_t base);
    void GUD900_drawImage(unsigned x, uint8_t y, unsigned width, uint8_t height, const uint8_t *data);
    void GUD900_drawImage(unsigned x, uint8_t y, ImageMemoryArea area, unsigned long address, uint8_t srcHeight, unsigned width, uint8_t height, unsigned offsetx, unsigned offsety);
    void GUD900_drawImage(unsigned x, uint8_t y, ImageMemoryArea area, unsigned long address, unsigned width, uint8_t height);
    
    void print_p(const char *str);
    void print_p(unsigned x, uint8_t y, const char *str);
    void print_p(unsigned x, uint8_t y, const char *buffer, uint8_t len);
    void GUD900_drawImage_p(unsigned width, uint8_t height, const uint8_t *data);
    void GUD900_drawImage_p(unsigned x, uint8_t y, unsigned width, uint8_t height, const uint8_t *data);
	
	//Touch functions
	void GUD900_touch_status_read_all();
	void GUD900_touch_status_read(uint8_t switchNum);
	void GUD900_touch_set(uint8_t mode);
	void GUD900_touch_level_read();
	void GUD900_touch_change_param(uint8_t mode, uint8_t value);
};
#endif